/*
 * Penmount driver for Fedora 9
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <X11/keysym.h>
#include <X11/XF86keysym.h>
#include <X11/extensions/XIproto.h>

#include <linux/input.h>
#include <linux/kd.h>

#include <sys/time.h>

#include <misc.h>
#include <xf86.h>
#include <xf86str.h>
#include <xf86_OSproc.h>
#include <xf86Xinput.h>
#include <xisb.h>
#include <exevents.h>
#include <mipointer.h>
#include <xf86Module.h>
#include <errno.h>
#include <fcntl.h>

/* 2.4 compatibility */
#ifndef EVIOCGRAB
#define EVIOCGRAB _IOW('E', 0x90, int)
#endif

#ifndef BTN_TASK
#define BTN_TASK 0x117
#endif

#ifndef EV_SYN
#define EV_SYN EV_RST
#endif
/* end compat */

#define ArrayLength(a) (sizeof(a) / (sizeof((a)[0])))

/* touch_flags */
#define DOING_CALIB             0x80
#define BUTTON_RIGHT_PRESSED    0x40
#define BUTTON_MIDDLE_PRESSED   0x20
#define BUTTON_LEFT_PRESSED     0x10
#define LOCK_WINDOW             0x08
#define ISSUE_2ND_BUTTON        0x04
#define PEN_UP                  0x02
#define PEN_DOWN                0x01

/*===========================================================================*/
typedef struct {
    int kernel24;
    int screen;                     /* screen no. */
    int min_x, min_y, max_x, max_y; /* for raw data range */
    int abs_x, abs_y;               /* raw data*/
    int flags;
    int xmin_offset, xmax_offset;   /* for shrink calib, edge offset */
    int ymin_offset, ymax_offset;
    int calib_mode;                 /* 0=standard calib, 4,9,16,25=advanced calib */
    int portrait;
    int std_xmin, std_xmax;         /* for standard calib range */
    int std_ymin, std_ymax;
    int screen_width, screen_height;      /* screen resolution */
    int screen_scale;               /* panel-->scr or panel-->raw data range*/
    int rotation;                   /* rotation */
    int pendown, old_pendown;       /* pendown or penup */
    int touch_flags;
    int calib_x, calib_y;           /* for calibration */
    int old_x, old_y;               /* previous abs_x, abs_y */
    int hold_x, hold_y;             /* press and hold, 1st point (pen-down) */
    int beep;
    int press_pitch, press_vol, press_dur;
    int rel_pitch, rel_vol, rel_dur;
    int average_point, mcnt;
    int av_x[8], av_y[8];
    int calib_hold_time, hold_time;
    int lock_window_range;
    int button2;
    int pendown_mode;
    int ad_bit;
    char configfile[64];
    int chip;
    unsigned char packet[6];
    int lex_mode;
    XISBuffer   *buffer;        /* raw data buffer */
    int proximity;
    int reporting_mode;
} PenmountRec, *PenmountPtr;

static const char *penmountDefaults[] = {
    "BaudRate",     "19200",
    "StopBits",     "1",
    "DataBits",     "8",
    "Parity",       "None",
    "Vmin",         "3",
    "Vtime",        "1",
    "FlowControl",  "None",
    NULL
};

#define GET_TIME(tv) (gettimeofday(&tv, (struct timezone *)NULL))
#define DIF_TIME(t1,t2) ((t2.tv_sec -t1.tv_sec) *1000000+ \
                           (t2.tv_usec-t1.tv_usec))
static struct timeval tv, tv0, tv1; /* time of last up, and down events */

typedef enum {
    PENMOUNT_BYTE0,
    PENMOUNT_BYTE1,
    PENMOUNT_BYTE2,
    PENMOUNT_BYTE3,
    PENMOUNT_BYTE4,
    PENMOUNT_BYTE5,
    PENMOUNT_RESPONSE0,
    PENMOUNT_RESPONSE1,
    PENMOUNT_RESPONSE2,
    PENMOUNT_RESPONSE3,
    PENMOUNT_RESPONSE4,
    PENMOUNT_RESPONSE5
} PenmountState;

typedef enum {
    UNKNOWN_CHIP = 0,
    PM9000,
    PM5000,
    PM6000USB,
    PM6000COM,
    PMLPC
} PenmountChip;

void LoadStdCalib(PenmountPtr pPenmount);


/*===========================================================================*/
/* 
 * The ConvertProc function may need to be tailored for your device.
 * This function converts the device's valuator outputs to x and y coordinates
 * to simulate mouse events.
 */
static Bool PenmountConvertProc (InputInfoPtr pInfo,
			 int first,
			 int num,
			 int v0,
			 int v1,
			 int v2,
			 int v3,
			 int v4,
			 int v5,
			 int *x,
			 int *y)
{
    PenmountPtr pPenmount = pInfo->private;

	if (pPenmount->reporting_mode == TS_Raw) {
        *x = xf86ScaleAxis (v0, 0, pPenmount->screen_width, pPenmount->min_x, pPenmount->max_x);
        *y = xf86ScaleAxis (v1, 0, pPenmount->screen_height, pPenmount->min_y, pPenmount->max_y);
    } else {
        *x = v0;
        *y = v1;
    }
	return (TRUE);
}
/*===========================================================================*/
/* The SwitchMode function may need to be tailored for your device */
static int PenmountSwitchMode (ClientPtr client, DeviceIntPtr dev, int mode)
{
    InputInfoPtr pInfo = dev->public.devicePrivate;
    PenmountPtr pPenmount = pInfo->private;

    if ((mode == TS_Raw) || (mode == TS_Scaled)) {
        pPenmount->reporting_mode = mode;
        return (Success);
    } else {
        return (!Success);
    }
}
/*===========================================================================*/
/* the CloseProc should not need to be tailored to your device */
static void PenmountCloseProc (InputInfoPtr pInfo)
{

}
/*===========================================================================*/
static int PenmountControlProc (InputInfoPtr pInfo, xDeviceCtl * control)
{
    xDeviceAbsCalibCtl *c = (xDeviceAbsCalibCtl *) control;
    PenmountPtr pPenmount = pInfo->private;

    pPenmount->min_x = c->min_x;
    pPenmount->max_x = c->max_x;
    pPenmount->min_y = c->min_y;
    pPenmount->max_y = c->max_y;

    return (Success);
}
/*===========================================================================*/
static Bool PenmountSendPacket (PenmountPtr pPenmount, unsigned char *buf, int len)
{
    int count = 0;

    while ( len > 0 ) {
        if ( XisbWrite(pPenmount->buffer, buf, 1) == 1 ) {
            buf++;
            len--;
			continue;
        }

        if ( count++ > 500 )
            break;
    }

    return (len ? !Success : Success);
}
/*===========================================================================*/
static Bool PenmountGetPacket (PenmountPtr pPenmount)
{
    int     cnt;
    int     c;

    cnt = 0;
    while ((c = XisbRead(pPenmount->buffer)) >= 0) {

        if (cnt++ > 500)
            return !Success;

        switch (pPenmount->lex_mode) {
            case PENMOUNT_BYTE0:
                if (pPenmount->chip == PM9000) {
                    if ( c != 0xFF && c != 0xBF )
                        return (!Success);
                } else if (pPenmount->chip == PM6000COM ) {
                    if ( c != 0x70 && c != 0x30 )
                        return (!Success);
                }
                pPenmount->packet[0] = (unsigned char) c;
                pPenmount->lex_mode  = PENMOUNT_BYTE1;
                break;
            case PENMOUNT_BYTE1:
                pPenmount->packet[1] = (unsigned char) c;
                pPenmount->lex_mode  = PENMOUNT_BYTE2;
                break;
            case PENMOUNT_BYTE2:
                pPenmount->packet[2] = (unsigned char) c;
                pPenmount->lex_mode  = PENMOUNT_BYTE3;
                break;
            case PENMOUNT_BYTE3:
                pPenmount->packet[3] = (unsigned char) c;
                pPenmount->lex_mode  = PENMOUNT_BYTE4;
                break;
            case PENMOUNT_BYTE4:
                pPenmount->packet[4] = (unsigned char) c;
                if (pPenmount->chip == PM9000) {
                    pPenmount->lex_mode  = PENMOUNT_BYTE0;
                    return Success;
                } else if (pPenmount->chip == PM6000COM) {
                    pPenmount->lex_mode  = PENMOUNT_BYTE5;
                }
                break;
            case PENMOUNT_BYTE5:
                pPenmount->packet[5] = (unsigned char) c;
                pPenmount->lex_mode  = PENMOUNT_BYTE0;
                return Success;
                break;
            default:
                break;
        }
    }

    return !Success;
}
/*===========================================================================*/
void IOWorkerThread(PenmountPtr pPenmount)
{
    int fd;

    if (access("/tmp/penmount_calib_ok", F_OK) == 0) { /* exist */

        pPenmount->touch_flags &= ~DOING_CALIB;

        remove("/tmp/penmount_calib_ok");
        remove("/tmp/penmount_calib");
        remove("/tmp/penmount_data");

        pPenmount->calib_mode = SetCalibration(pPenmount->configfile, pPenmount->rotation);
        if (pPenmount->calib_mode == 0)
            LoadStdCalib(pPenmount);

        pPenmount->min_x = 0;
        pPenmount->min_y = 0;
        pPenmount->max_x = 2047;
        pPenmount->max_y = 2047;
    }

    if (access("/tmp/penmount_calib", F_OK) == 0) { /* exist */
        pPenmount->touch_flags |= DOING_CALIB;
    }

}
/*===========================================================================*/
static void ProcessPenUp(InputInfoPtr pInfo)
{
    PenmountPtr pPenmount = pInfo->private;

    /*f86Msg(X_INFO, "ProcessPenUp\n");*/

    if (pPenmount->touch_flags & PEN_UP) {
        pPenmount->touch_flags &= ~PEN_UP;

        if (pPenmount->touch_flags & BUTTON_LEFT_PRESSED) {
            /* LEFT pressed */
            pPenmount->touch_flags &= ~BUTTON_LEFT_PRESSED;
            if (pPenmount->touch_flags & LOCK_WINDOW) 
                xf86PostButtonEvent (pInfo->dev, 0, 1, 0, 0, 0);
            else
                xf86PostButtonEvent (pInfo->dev, 0, 1, 0, 0, 0);   /* LEFT released*/
        }

        if (pPenmount->touch_flags & BUTTON_RIGHT_PRESSED) {
            /* RIGHT pressed */
            pPenmount->touch_flags &= ~BUTTON_RIGHT_PRESSED;
            xf86PostButtonEvent (pInfo->dev, 0, 3, 0, 0, 0);   /* RIGHT released */
        } else if (pPenmount->touch_flags & BUTTON_MIDDLE_PRESSED) {
            /* MIDDLE pressed */
            pPenmount->touch_flags &= ~BUTTON_MIDDLE_PRESSED;
            xf86PostButtonEvent (pInfo->dev, 0, 2, 0, 0, 0);   /* MIDDLE released */
        }

        pPenmount->hold_x = pPenmount->old_x = 0;
        pPenmount->hold_y = pPenmount->old_y = 0;
        pPenmount->touch_flags &= ~LOCK_WINDOW;

    }
}
/*===========================================================================*/
static void ProcessPenDown(InputInfoPtr pInfo)
{
    PenmountPtr pPenmount = pInfo->private;

    /*xf86Msg(X_INFO, "ProcessPenDown\n");*/

    if (pPenmount->touch_flags & PEN_DOWN) {
        pPenmount->touch_flags &= ~PEN_DOWN;

        if (pPenmount->pendown_mode == 0) {
            /* stream mode */
            pPenmount->touch_flags |= BUTTON_LEFT_PRESSED;
            xf86PostButtonEvent (pInfo->dev, 0, 1, 1, 0, 0);   /* LEFT pressed */
        } else {
            /* point mode */
            xf86PostButtonEvent (pInfo->dev, 0, 1, 1, 0, 0);   /* LEFT pressed */
            xf86PostButtonEvent (pInfo->dev, 0, 1, 0, 0, 0);   /* LEFT released*/
        }
    }
}
/*===========================================================================*/
int CreatePenmountData(PenmountPtr pPenmount)
{
    FILE    *f;
    int     mode;

    mode = F_OK;
    if (access("/tmp/penmount_calib", mode) != 0) {
        if (access("/tmp/penmount_data", mode) != 0)
            remove("/tmp/penmount_data");
        return -1;
    }

    f = fopen("/tmp/penmount_data", "w");
    if (!f)
        return -1;  /* file not found */

    fprintf(f, "%d ", pPenmount->calib_x);
    fprintf(f, "%d ", pPenmount->calib_y);
    fprintf(f, "%d ", pPenmount->rotation);
    fclose(f);

}
/*===========================================================================*/
static void Issue2ndButton(InputInfoPtr pInfo)
{
    PenmountPtr pPenmount = pInfo->private;


    if (pPenmount->touch_flags & ISSUE_2ND_BUTTON) {
        pPenmount->touch_flags &= ~(ISSUE_2ND_BUTTON);

        /*xf86Msg(X_INFO, "Issue2ndButton\n");*/

        /* prior to send post button event */
        CreatePenmountData(pPenmount);

        /* Send Left released before send RIGHT/MIDDLE pressed */
        xf86PostButtonEvent (pInfo->dev, 0, 1, 0, 0, 0);   /* LEFT released */
        pPenmount->touch_flags &= ~BUTTON_LEFT_PRESSED;

        if (pPenmount->button2 == 3) {
            pPenmount->touch_flags |= BUTTON_RIGHT_PRESSED;
            xf86PostButtonEvent (pInfo->dev, 0, 3, 1, 0, 0);   /* RIGHT pressed */
        } else {
            pPenmount->touch_flags |= BUTTON_MIDDLE_PRESSED;
            xf86PostButtonEvent (pInfo->dev, 0, 2, 1, 0, 0);   /* MIDDLE pressed */
        }
    }
}
/*===========================================================================*/
static void PressAndHold(PenmountPtr pPenmount)
{
    int temp;

    /*xf86Msg(X_INFO, "PressAndHold: start\n");*/

    if (pPenmount->touch_flags & LOCK_WINDOW) {
        /*xf86Msg(X_INFO, "PressAndHold: get time\n");*/
        /*GET_TIME(tv1);*/
        if (pPenmount->chip == PM6000USB ||
            pPenmount->chip == PM5000    ||
            pPenmount->chip == PMLPC) {
            tv1.tv_sec  = tv.tv_sec;
            tv1.tv_usec = tv.tv_usec;
        } else if (pPenmount->chip == PM6000COM || pPenmount->chip == PM9000) {
            GET_TIME(tv1);
        }

        temp = pPenmount->calib_x + pPenmount->old_x;
        pPenmount->calib_x = temp / 2;
        temp = pPenmount->calib_y + pPenmount->old_y;
        pPenmount->calib_y = temp / 2;

        if (pPenmount->touch_flags & DOING_CALIB) {
            /* calib hold time > 2000ms */
            if (DIF_TIME(tv0, tv1) > pPenmount->calib_hold_time) {
                pPenmount->touch_flags |= ISSUE_2ND_BUTTON;
                pPenmount->touch_flags &= ~LOCK_WINDOW;
                /*xf86Msg(X_INFO, "PressAndHold: issue 2nd 2\n");*/
            }
            /* exceed range */
            if (abs(pPenmount->hold_x - pPenmount->old_x) >= pPenmount->lock_window_range ||
                abs(pPenmount->hold_y - pPenmount->old_y) >= pPenmount->lock_window_range)
                pPenmount->touch_flags &= ~LOCK_WINDOW;
        } else {
            /* hold time > 700ms */
            if (DIF_TIME(tv0, tv1) > pPenmount->hold_time) {
                pPenmount->touch_flags |= ISSUE_2ND_BUTTON;
                pPenmount->touch_flags &= ~LOCK_WINDOW;
                /*xf86Msg(X_INFO, "PressAndHold: issue 2nd\n");*/
            }
            /* exceed range */
            if (abs(pPenmount->hold_x - pPenmount->old_x) >= pPenmount->lock_window_range ||
                abs(pPenmount->hold_y - pPenmount->old_y) >= pPenmount->lock_window_range)
                pPenmount->touch_flags &= ~LOCK_WINDOW;
        }
    }

    /*xf86Msg(X_INFO, "PressAndHold: end\n");*/

}

/*===========================================================================*/
static void AveragePoint(PenmountPtr pPenmount, int *x, int *y)
{
    int ax, ay;
    int i;


    switch(pPenmount->average_point) {
    case 4:
        pPenmount->av_x[pPenmount->mcnt%4] = *x;
        pPenmount->av_y[pPenmount->mcnt%4] = *y;
        if (pPenmount->mcnt < 4) {
            ax = *x;
            ay = *y;
        } else {
            for (i=0, ax=0, ay=0; i<4; i++) {
                ax += pPenmount->av_x[i];
                ay += pPenmount->av_y[i];
            }
            ax /= 4;
            ay /= 4;
        }
        break;

    case 8:
        pPenmount->av_x[pPenmount->mcnt%8] = *x;
        pPenmount->av_y[pPenmount->mcnt%8] = *y;
        if (pPenmount->mcnt < 8) {
            ax = *x;
            ay = *y;
        } else {
            for (i=0, ax=0, ay=0; i<8; i++) {
                ax += pPenmount->av_x[i];
                ay += pPenmount->av_y[i];
            }
            ax /= 8;
            ay /= 8;
        }
        break;

    default:
        ax = *x;
        ay = *y;
        break;
    }/* switch */

    pPenmount->mcnt++;

    *x = ax;
    *y = ay;
}
/*===========================================================================*/
static int sys_beep(int freq, int vol, int dur)
{
    int     fd;
    int     f;
    
    if (freq < 20)
        f = 20;
    else if (freq > 32767)
        f = 32767;
    else
        f = freq;
    
    fd = open("/dev/console", O_WRONLY);
    if (!fd) {
        return -1;
    }

    ioctl(fd, KIOCSOUND, (int)(1193180/f));
    usleep(vol*dur*20);
    ioctl(fd, KIOCSOUND, 0);
    
    close(fd);
    
}

/* cannot call xf86SoundKbdBell @ XINPUT_ABI_VERSIOM <> 0 */    
/*xf86SoundKbdBell(priv->press_vol, priv->press_pitch, priv->press.dur);*/
/*===========================================================================*/
static void Beeper(PenmountPtr pPenmount, int press)
{
    if (pPenmount->beep) {
        if (press) {
            sys_beep(pPenmount->press_pitch, pPenmount->press_vol, pPenmount->press_dur);
        } else {
            sys_beep(pPenmount->rel_pitch, pPenmount->rel_vol, pPenmount->rel_dur);
        }
    }
}
/*===========================================================================*/
static void PenDownUpMoment(PenmountPtr pPenmount)
{
    ScrnInfoPtr pScrn;
    Rotation rotation;

    /* get pen-down moment */
    if (pPenmount->old_pendown == 0 && pPenmount->pendown == 1) {
        if (pPenmount->chip == PM6000USB || pPenmount->chip == PM5000 || pPenmount->chip == PMLPC) {
            tv0.tv_sec  = tv.tv_sec;
            tv0.tv_usec = tv.tv_usec;
        } else if (pPenmount->chip == PM6000COM || pPenmount->chip == PM9000) {
            GET_TIME(tv0);
        }

        pPenmount->touch_flags = 0;  /* clear touch_flags */

        /* hold X, Y (raw) */
        pPenmount->hold_x = pPenmount->abs_x;
        pPenmount->hold_y = pPenmount->abs_y;
        pPenmount->old_x  = pPenmount->abs_x;
        pPenmount->old_y  = pPenmount->abs_y;

        pPenmount->touch_flags |= LOCK_WINDOW;
        pPenmount->touch_flags |= PEN_DOWN;
            
        Beeper(pPenmount, 1);
            
        pPenmount->mcnt = 0;

        /* FIFO operation */
        IOWorkerThread(pPenmount);

        pPenmount->calib_x = pPenmount->abs_x;
        pPenmount->calib_y = pPenmount->abs_y;

        /* ROTATION */
        pScrn = xf86Screens[pPenmount->screen];
        rotation = RRGetRotation(pScrn->pScreen);
        if (rotation != pPenmount->rotation) {
            pPenmount->rotation = rotation;
            pPenmount->calib_mode = SetCalibration(pPenmount->configfile, pPenmount->rotation);
            if (pPenmount->calib_mode == 0)
                LoadStdCalib(pPenmount);
        }
        pPenmount->rotation = rotation;

    /* get pen-up moment */
    } else if (pPenmount->old_pendown == 1 && pPenmount->pendown == 0) {
        pPenmount->touch_flags |= PEN_UP;
        Beeper(pPenmount, 0);
    }
}
/*===========================================================================*/
/*
 * standard calibration /etc/penmount.dat
 
 * <CalibPointsNum>  0
 * <Top Point>       64 64
 * <Right Point>     1984 64
 * <Bottom Point>    1984 1984
 * <Left Point>      64 1984
 * <Rotate>          1|2|4|8
 */
void LoadStdCalib(PenmountPtr pPenmount)
{
    FILE *f = fopen(pPenmount->configfile, "r");
    int     px[4], py[4];
    int     tmpx[4], tmpy[4];
    int     i, rotate;
    int     rotate_table[4][8] = {
            { 1, 2, 3, 4, 5, 6, 7, 8 },  /* normal   */
            { 7, 5, 8, 6, 3, 1, 4, 2 },  /* left     */
            { 4, 3, 2, 1, 8, 7, 6, 5 },  /* inverted */
            { 6, 8, 5, 7, 2, 4, 1, 3 }}; /* right    */


    if(!f) {
        /* avoid bad value cause the system crash */    
        pPenmount->std_xmin = 20;
        pPenmount->std_ymin = 20;
        pPenmount->std_xmax = 2000;
        pPenmount->std_ymax = 2000;
        pPenmount->portrait = 1;
        return;
    }
    /*
         +--0------------+
         |               |
         |               1
         |               |
         3               |
         |               |
         +------------2--+
     */

    fscanf(f, "%d", &i);  /* useless */
    for (i=0; i<4; i++) {
        fscanf(f, "%d", &px[i]);
        fscanf(f, "%d", &py[i]);
    }
    fscanf(f, "%d", &rotate);
    fclose(f);

    for (i=0; i<4; i++) {
        tmpx[i] = px[i];
        tmpy[i] = py[i];
    }    

    /* rotation */
    switch(rotate) {
    case 1: /* normal */
        break;
    case 2: /* left */
        px[0] = tmpx[1]; py[0] = tmpy[1];
        px[1] = tmpx[2]; py[1] = tmpy[2];
        px[2] = tmpx[3]; py[2] = tmpy[3];
        px[3] = tmpx[0]; py[3] = tmpy[0];
        break;
    case 4: /* inverted */
        px[0] = tmpx[2]; py[0] = tmpy[2];
        px[1] = tmpx[3]; py[1] = tmpy[3];
        px[2] = tmpx[0]; py[2] = tmpy[0];
        px[3] = tmpx[1]; py[3] = tmpy[1];
        break;
    case 8: /* right */
        px[0] = tmpx[3]; py[0] = tmpy[3];
        px[1] = tmpx[0]; py[1] = tmpy[0];
        px[2] = tmpx[1]; py[2] = tmpy[1];
        px[3] = tmpx[2]; py[3] = tmpy[2];
        break;
    }

    /* determine portrait mode */
    if ( (px[3]<px[0] && px[0]<px[2] && px[2]<px[1]) &&
         (py[0]<py[1] && py[1]<py[3] && py[3]<py[2]) )
        pPenmount->portrait = 1;
    else if ( (px[3]<px[0] && px[0]<px[2] && px[2]<px[1]) &&
              (py[2]<py[3] && py[3]<py[1] && py[1]<py[0]) )
        pPenmount->portrait = 2;
    else if ( (px[1]<px[2] && px[2]<px[0] && px[0]<px[3]) &&
              (py[0]<py[1] && py[1]<py[3] && py[3]<py[2]) )
        pPenmount->portrait = 3;
    else if ( (px[1]<px[2] && px[2]<px[0] && px[0]<px[3]) &&
              (py[2]<py[3] && py[3]<py[1] && py[1]<py[0]) )
        pPenmount->portrait = 4;
    else if ( (px[0]<px[1] && px[1]<px[3] && px[3]<px[2]) &&
              (py[3]<py[0] && py[0]<py[2] && py[2]<py[1]) )
        pPenmount->portrait = 5;
    else if ( (px[2]<px[3] && px[3]<px[1] && px[1]<px[0]) &&
              (py[3]<py[0] && py[0]<py[2] && py[2]<py[1]) )
        pPenmount->portrait = 6;
    else if ( (px[0]<px[1] && px[1]<px[3] && px[3]<px[2]) &&
              (py[1]<py[2] && py[2]<py[0] && py[0]<py[3]) )
        pPenmount->portrait = 7;
    else if ( (px[2]<px[3] && px[3]<px[1] && px[1]<px[0]) &&
              (py[1]<py[2] && py[2]<py[0] && py[0]<py[3]) )
        pPenmount->portrait = 8;
    else
        pPenmount->portrait = 1;

    /* determine xmin/xmax, ymin/ymax */
    pPenmount->std_xmin = pPenmount->std_ymin = 65535;
    pPenmount->std_xmax = pPenmount->std_ymax = 0;
    for (i=0; i<4; i++) {
        if (px[i] > pPenmount->std_xmax)
            pPenmount->std_xmax = px[i];
        if (px[i] < pPenmount->std_xmin)
            pPenmount->std_xmin = px[i];
        if (py[i] > pPenmount->std_ymax)
            pPenmount->std_ymax = py[i];
        if (py[i] < pPenmount->std_ymin)
            pPenmount->std_ymin = py[i];
    }

    if (pPenmount->rotation == 1)
        rotate = 0;
    else if (pPenmount->rotation == 2)
        rotate = 1;
    else if (pPenmount->rotation == 4)
        rotate = 2;
    else if (pPenmount->rotation == 8)
        rotate = 3;
    else
        rotate = 0;

    i = (pPenmount->portrait - 1);    /* original pmode 1~8 --> 0~7 */
    pPenmount->portrait = rotate_table[rotate][i];

}

/*===========================================================================*/
static void DoStdCalib(PenmountPtr pPenmount, int *px, int *py)
{
    int pmode = pPenmount->portrait;
    int x, y, temp;

    x = *px;
    y = *py;

    if (pmode >= 5 && pmode <= 8) { /* swap x, y */
        temp = x;
        x    = y;
        y    = temp;
    }

    if (x < pPenmount->std_xmin)
        x = pPenmount->std_xmin;
    if (x > pPenmount->std_xmax)
        x = pPenmount->std_xmax;
    if (y < pPenmount->std_ymin)
        y = pPenmount->std_ymin;
    if (y > pPenmount->std_ymax)
        y = pPenmount->std_ymax;

    
    if (pmode == 1 || pmode == 2 || pmode == 5 || pmode == 6) {
        if (pPenmount->screen_scale == 1)
            x = ((x - pPenmount->std_xmin) * pPenmount->screen_width) / 
                (pPenmount->std_xmax - pPenmount->std_xmin);
        else
            x = ((x - pPenmount->std_xmin) * 2047) /
                (pPenmount->std_xmax - pPenmount->std_xmin);
    } else {
        if (pPenmount->screen_scale == 1)
            x = ((pPenmount->std_xmax - x) * pPenmount->screen_width) /
                (pPenmount->std_xmax - pPenmount->std_xmin);
        else
            x = ((pPenmount->std_xmax - x) * 2047) /
                (pPenmount->std_xmax - pPenmount->std_xmin);
    }

    if (pmode == 1 || pmode == 3 || pmode == 5 || pmode == 7) {
        if (pPenmount->screen_scale == 1)
            y = ((y - pPenmount->std_ymin) * pPenmount->screen_height) /
                (pPenmount->std_ymax - pPenmount->std_ymin);
        else
            y = ((y - pPenmount->std_ymin) * 2047) /
                (pPenmount->std_ymax - pPenmount->std_ymin);
    } else {
        if (pPenmount->screen_scale == 1)
            y = ((pPenmount->std_ymax - y) * pPenmount->screen_height) /
                (pPenmount->std_ymax - pPenmount->std_ymin);
        else
            y = ((pPenmount->std_ymax - y) * 2047) /
                (pPenmount->std_ymax - pPenmount->std_ymin);
    }
    
    *px = x;
    *py = y;
}
/*===========================================================================*/
static void Penmount232ReadInput(InputInfoPtr pInfo)
{
    PenmountPtr pPenmount = pInfo->private;
    int x, y;

    pPenmount->screen_width  = screenInfo.screens[pPenmount->screen]->width;
    pPenmount->screen_height = screenInfo.screens[pPenmount->screen]->height;

    XisbBlockDuration (pPenmount->buffer, -1);

    while (1) {
        if (PenmountGetPacket(pPenmount) != Success)
            break;

        /*
        xf86Msg(X_INFO, "232ReadInput: %02x %02x %02x %02x %02x\n",
                pPenmount->packet[0],
                pPenmount->packet[1],
                pPenmount->packet[2],
                pPenmount->packet[3],
                pPenmount->packet[4]);
        */

        /* forced to 10-bit */
        /*
        if (pPenmount->ad_bit == 10 && (pPenmount->packet[2] > 0x03 || pPenmount->packet[4] > 0x03))
            pPenmount->ad_bit = 12;
        */


        if (pPenmount->chip == PM9000) {
            if (pPenmount->packet[0] == 0xFF) {
                pPenmount->old_pendown = pPenmount->pendown;
                pPenmount->pendown = 1; /* pen down */
            } else if (pPenmount->packet[0] == 0xBF ) {
                pPenmount->old_pendown = pPenmount->pendown;
                pPenmount->pendown = 0; /* pen up */
            }

            /* 9000 : FF/BF, XH, XL, YH, YL (bit7 unused) */
            x = pPenmount->packet[1] & 0x07;    /* -----987 */
            x <<= 7;                            /* ------98 7------- */
            x += pPenmount->packet[2];          /* ------98 76543210 */
            y = pPenmount->packet[3] & 0x07;    /* -----987 */
            y <<= 7;                            /* ------98 7------- */
            y += pPenmount->packet[4];          /* ------98 76543210 */

            x <<= 1;    /* convert 11-bit */
            y <<= 1;

        } else if (pPenmount->chip == PM6000COM) {
            /* none checksum */
            if ( pPenmount->packet[0] == 0x70 ) {
                pPenmount->old_pendown = pPenmount->pendown;
                pPenmount->pendown = 1; /* pen down */
            } else if ( pPenmount->packet[0] == 0x30 ) {
                pPenmount->old_pendown = pPenmount->pendown;
                pPenmount->pendown = 0; /* pen up */
            }
     
            /* 6000 : 70/30, XL, XH, YL, YH */
            if (pPenmount->ad_bit == 12) {
                x = ( pPenmount->packet[2] & 0x0f ); /* ----BA98 */
                x <<= 8;                             /* ----BA98 -------- */
                x += pPenmount->packet[1];           /* ----BA98 76543210 */
                y = ( pPenmount->packet[4] & 0x0f ); /* ----BA98 */
                y <<= 8;                             /* ----BA98 -------- */
                y += pPenmount->packet[3];           /* ----BA98 76543210 */

                x >>= 1;                        /* -----BA8 87654321 , 11-bit */
                y >>= 1;                        /* -----BA8 87654321 , 11-bit */
            } else {
                x = ( pPenmount->packet[2] & 0x03 ); /* ------98 */
                x <<= 8;                             /* ------98 -------- */
                x += pPenmount->packet[1];           /* ------98 76543210 */
                y = ( pPenmount->packet[4] & 0x03 ); /* ------98 */
                y <<= 8;                             /* ------98 -------- */
                y += pPenmount->packet[3];           /* ------98 76543210 */

                x <<= 1;                        /* -----987 6543210x , 11-bit */
                y <<= 1;                        /* -----987 6543210x , 11-bit */
            }
        }

        pPenmount->old_x = pPenmount->abs_x;
        pPenmount->old_y = pPenmount->abs_y;
        pPenmount->abs_x = x;           /* convert to 11-bit */
        pPenmount->abs_y = y;           /* convert to 11-bit */

        xf86Msg(X_INFO, "(%d, %d)-->", x, y);

        PenDownUpMoment(pPenmount);

        ProcessPenDown(pInfo);

        ProcessPenUp(pInfo);

        if (pPenmount->calib_mode == 0) { /* standard calib */
            DoStdCalib(pPenmount, &x, &y);
        } else {
            DoCalibrate(pPenmount->xmin_offset,
                        pPenmount->xmax_offset,
                        pPenmount->ymin_offset,
                        pPenmount->ymax_offset,
                        &x, &y);

            if (pPenmount->screen_scale == 1) {
        	    x = (x * pPenmount->screen_width)  / (pPenmount->max_x - pPenmount->min_x);
	            y = (y * pPenmount->screen_height) / (pPenmount->max_y - pPenmount->min_y);
                xf86Msg(X_INFO, "scr(%d, %d) %dx%d\n", x, y,
                        pPenmount->screen_width, pPenmount->screen_height);
	        }

        }

		xf86XInputSetScreen (pInfo, pPenmount->screen, x, y);
        if ((pPenmount->proximity == FALSE) && (pPenmount->pendown == 1)) {
            pPenmount->proximity = TRUE;
            xf86PostProximityEvent (pInfo->dev, 1, 0, 2, x, y);
        }

        AveragePoint(pPenmount, &x, &y);

        xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, x, y);

        PressAndHold(pPenmount);
        
        Issue2ndButton(pInfo);

        if ((pPenmount->proximity == TRUE) && (pPenmount->pendown == 0)) {
            pPenmount->proximity = FALSE;
            xf86PostProximityEvent(pInfo->dev, 0, 0, 2, x, y);
        }


    }/* while */
}
/*===========================================================================*/
static void PenmountReadInput(InputInfoPtr pInfo)
{
    struct input_event ev;
    int len, value;
    int x, y;
    unsigned int abs;
    PenmountPtr pPenmount = pInfo->private;
    int scr_w = screenInfo.screens[pPenmount->screen]->width;
    int scr_h = screenInfo.screens[pPenmount->screen]->height;
    
    pPenmount->screen_width  = scr_w;
    pPenmount->screen_height = scr_h;

    abs = 0;

    /*xf86Msg(X_INFO, "penmount: ReadInput\n");*/
    while (xf86WaitForInput (pInfo->fd, 0) > 0) {
        len = read(pInfo->fd, &ev, sizeof ev);
        if (len != sizeof ev) {
            /* The kernel promises that we always only read a complete
             * event, so len != sizeof ev is an error. */
		    value *= 2;
            xf86Msg(X_ERROR, "Read error: %s\n", strerror(errno));
            break;
        }

        /* Get the signed value, earlier kernels had this as unsigned */
        value = ev.value;

        /* Get timestamp */
        tv.tv_sec  = ev.time.tv_sec;
        tv.tv_usec = ev.time.tv_usec;

    	/*xf86Msg(X_INFO, " %04x:%04x=%d\n", ev.type, ev.code, value);*/

        switch (ev.type) {
        case EV_ABS:
            switch (ev.code) {
            case ABS_X:
                if (pPenmount->ad_bit == 12)
                    value >>= 1;
                else
        		    value *= 2; /* convert to 11-bit */
        		pPenmount->old_x = pPenmount->abs_x;
                pPenmount->abs_x = value;
                abs = 1;
                break;
            case ABS_Y:
                if (pPenmount->ad_bit == 12)
                    value >>= 1;
                else
            		value *= 2; /* convert to 11-bit */
        		pPenmount->old_y = pPenmount->abs_y;
                pPenmount->abs_y = value;
                abs = 1;
                break;
            }
            break;

        case EV_KEY:
            if (ev.code >= BTN_MOUSE && ev.code < KEY_OK)
                if (value == 2)
                    break;
            if (ev.code == BTN_LEFT) {
                pPenmount->old_pendown = pPenmount->pendown;
                pPenmount->pendown = value;
                /*xf86PostButtonEvent(pInfo->dev, 0, 1, value, 0, 0);*/
                PenDownUpMoment(pPenmount);
                ProcessPenDown(pInfo);
                ProcessPenUp(pInfo);
            }
            break;

        case EV_SYN:
        case EV_MSC:
            abs = 1;    /* keep press and hold better */
            break;
        default:
            break;
        }
    }/* while */

    /*
     * Some devices only generate valid abs coords when BTN_DIGI is
     * pressed.  On wacom tablets, this means that the pen is in
     * proximity of the tablet.  After the pen is removed, BTN_DIGI is
     * released, and a (0, 0) absolute event is generated.  Checking
     
     * initialized to 1 so devices that doesn't use this scheme still
     * just works.
     */
    if (abs) {

        /*xf86Msg(X_INFO, "penmount: (%d, %d)-->", pPenmount->abs_x, pPenmount->abs_y);*/
        x = pPenmount->abs_x;
        y = pPenmount->abs_y;


        if (pPenmount->calib_mode == 0) { /* standard calib */
            DoStdCalib(pPenmount, &x, &y);
        } else {
            DoCalibrate(pPenmount->xmin_offset,
                        pPenmount->xmax_offset,
                        pPenmount->ymin_offset,
                        pPenmount->ymax_offset,
                        &x, &y);

            if (pPenmount->screen_scale == 1) {
        	    x = (x * scr_w) / (pPenmount->max_x - pPenmount->min_x);
	            y = (y * scr_h) / (pPenmount->max_y - pPenmount->min_y);
                /*xf86Msg(X_INFO, "scr(%d, %d) %dx%d\n", x, y, scr_w, scr_h);*/
	        }

        }
        /*xf86Msg(X_INFO, "(%d, %d)-->\n", x, y);*/

        AveragePoint(pPenmount, &x, &y);

        PressAndHold(pPenmount);
        
        Issue2ndButton(pInfo);

        xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, x, y);
    }
}
/*===========================================================================*/
static void PenmountPtrCtrlProc(DeviceIntPtr device, PtrCtrl *ctrl)
{
    /* Nothing to do, dix handles all settings */
}
/*===========================================================================*/
static int PenmountAbsClass(DeviceIntPtr device)
{
    InputInfoPtr pInfo;
    PenmountPtr pPenmount;
    struct input_absinfo absinfo_x, absinfo_y;

    pInfo = device->public.devicePrivate;
    pPenmount = pInfo->private;

    if (pPenmount->chip == PM6000USB ||
        pPenmount->chip == PM5000 ||
        pPenmount->chip == PMLPC) {

        if (ioctl(pInfo->fd, EVIOCGABS(ABS_X), &absinfo_x) < 0) {
            xf86Msg(X_ERROR, "ioctl EVIOCGABS failed: %s\n", strerror(errno));
            return !Success;
        }

        if (ioctl(pInfo->fd, EVIOCGABS(ABS_Y), &absinfo_y) < 0) {
            xf86Msg(X_ERROR, "ioctl EVIOCGABS failed: %s\n", strerror(errno));
            return !Success;
        }

        pPenmount->min_x = absinfo_x.minimum;
        pPenmount->max_x = absinfo_x.maximum;
        pPenmount->min_y = absinfo_y.minimum;
        pPenmount->max_y = absinfo_y.maximum;
    }

    pPenmount->min_x = 0;
    pPenmount->max_x = 2047;
    pPenmount->min_y = 0;
    pPenmount->max_y = 2047;


    xf86Msg(X_INFO, "penmount: min_x=%d, max_x=%d, min_y=%d, max_y=%d\n", 
            pPenmount->min_x, pPenmount->max_x, pPenmount->min_y, pPenmount->max_y);


    if (pPenmount->chip == PM6000USB || pPenmount->chip == PM5000 || pPenmount->chip == PMLPC) {
        if (!InitValuatorClassDeviceStruct(device, 2, GetMotionHistory,
                                       GetMotionHistorySize(), Absolute))
            return !Success;
        /* X valuator */
        xf86InitValuatorAxisStruct(device, 0, pPenmount->min_x, pPenmount->max_x,
                       10000, 0, 10000);
        xf86InitValuatorDefaults(device, 0);

        /* Y valuator */
        xf86InitValuatorAxisStruct(device, 1, pPenmount->min_y, pPenmount->max_y,
                       10000, 0, 10000);
        xf86InitValuatorDefaults(device, 1);
        xf86MotionHistoryAllocate(pInfo);

        if (!InitPtrFeedbackClassDeviceStruct(device, PenmountPtrCtrlProc))
            return !Success;

    }

    if (pPenmount->chip == PM6000COM || pPenmount->chip == PM9000) {
    	if (InitValuatorClassDeviceStruct (device, 2, xf86GetMotionEvents,
	    				   pInfo->history_size, Absolute) == FALSE)
			return !Success;

        InitValuatorAxisStruct (device, 0, pPenmount->min_x, pPenmount->max_x,
            9500, 0, 9500);
        InitValuatorAxisStruct (device, 1, pPenmount->min_y, pPenmount->max_y,
            10500, 0, 10500);
		
        if (InitProximityClassDeviceStruct (device) == FALSE)
			return !Success;
        if (InitPtrFeedbackClassDeviceStruct(device, PenmountPtrCtrlProc) == FALSE)
			return !Success;
		
        xf86MotionHistoryAllocate (pInfo);
    }


    return Success;
}
/*===========================================================================*/
static int PenmountButtonClass(DeviceIntPtr device)
{
    /*CARD8 map[3];*/
    char map[] = {0, 1, 2};
    InputInfoPtr pInfo;
    int i;

    pInfo = device->public.devicePrivate;

    /* FIXME: count number of actual buttons */
    for (i = 0; i < ArrayLength(map); i++)
        map[i] = i;

    if (!InitButtonClassDeviceStruct(device, ArrayLength(map), map))
        return !Success;

    return Success;
}
/*===========================================================================*/
static int PenmountInit(DeviceIntPtr device)
{
    InputInfoPtr pInfo;
    PenmountPtr pPenmount;

    pInfo = device->public.devicePrivate;
    pPenmount = pInfo->private;

    PenmountButtonClass(device);
    
    PenmountAbsClass(device);

    return Success;
}
/*===========================================================================*/
static int PenmountProc(DeviceIntPtr device, int what)
{
    InputInfoPtr pInfo;
    PenmountPtr pPenmount;
    char enable_6000_cmd[6] = { 0xf1, 0, 0, 0, 0, 0x0e };
    char enable_9000_cmd[5] = { 0xf1, 0, 0, 0, 0 };


    /*xf86Msg(X_INFO, "penmount: Proc\n");*/

    pInfo = device->public.devicePrivate;
    pPenmount = pInfo->private;

    switch (what) {
    case DEVICE_INIT:
        return PenmountInit(device);

    case DEVICE_ON:
        if (pPenmount->chip == PM9000 || pPenmount->chip == PM6000COM) {

            pInfo->fd = xf86OpenSerial(pInfo->options);
            if (pInfo->fd == -1) {
                xf86Msg(X_WARNING, "%s: cannot open input device\n", pInfo->name);
                return (!Success);
            }

            pPenmount->buffer = XisbNew(pInfo->fd, 64);
            if (!pPenmount->buffer) {
                xf86CloseSerial(pInfo->fd);
                pInfo->fd = -1;
                return (!Success);
            }

            XisbBlockDuration (pPenmount->buffer, 500000);
            if (pPenmount->chip == PM6000COM) {
                PenmountSendPacket(pPenmount, enable_6000_cmd, 6);
                /*XisbWrite(pPenmount->buffer, enable_6000_cmd, 6);*/
                PenmountGetPacket(pPenmount);
            } else if (pPenmount->chip == PM9000){
                /*XisbWrite(pPenmount->buffer, enable_9000_cmd, 5);*/
                PenmountSendPacket(pPenmount, enable_9000_cmd, 5);
                if (PenmountGetPacket(pPenmount) == Success) {
                    /* set range to mininum */
                    enable_9000_cmd[0] = 0xf3;
                    /*XisbWrite(pPenmount->buffer, enable_9000_cmd, 5);*/
                    PenmountSendPacket(pPenmount, enable_9000_cmd, 5);
                }
                PenmountGetPacket(pPenmount);
            }

            XisbBlockDuration (pPenmount->buffer, -1);
            pPenmount->lex_mode = PENMOUNT_BYTE0;
            xf86FlushInput(pInfo->fd);

            AddEnabledDevice (pInfo->fd);

        } else { /* Penmount USB series */
            if (!pPenmount->kernel24 && ioctl(pInfo->fd, EVIOCGRAB, (void *)1))
                xf86Msg(X_WARNING, "%s: Grab failed (%s)\n", pInfo->name, strerror(errno));

            xf86AddEnabledDevice(pInfo);
        }
        
        device->public.on = TRUE;
        break;
        
    case DEVICE_OFF:
        xfree(pPenmount);/* ?? */
        if (pPenmount->chip == PM5000 || pPenmount->chip == PM6000USB) {
            if (!pPenmount->kernel24 && ioctl(pInfo->fd, EVIOCGRAB, (void *)0))
                xf86Msg(X_WARNING, "%s: Release failed (%s)\n", pInfo->name, strerror(errno));
        } else if (pPenmount->chip == PM9000 || pPenmount->chip == PM6000COM) {
            if (pInfo->fd != -1) {
                RemoveEnabledDevice (pInfo->fd);
                if (pPenmount->buffer) {
					XisbFree(pPenmount->buffer);
					pPenmount->buffer = NULL;
				}
			    xf86CloseSerial(pInfo->fd);
			}
	        device->public.on = FALSE;
    	    return (Success);
        }
        xf86RemoveEnabledDevice(pInfo);
        device->public.on = FALSE;
        break;

    case DEVICE_CLOSE:
        xf86Msg(X_INFO, "%s: Close\n", pInfo->name);
        if (pPenmount->chip == PM9000 || pPenmount->chip == PM6000COM) {
            if (pInfo->fd != -1) {
                RemoveEnabledDevice (pInfo->fd);
                if (pPenmount->buffer) {
					XisbFree(pPenmount->buffer);
					pPenmount->buffer = NULL;
				}
			    xf86CloseSerial(pInfo->fd);
			}
	        device->public.on = FALSE;
    	    return (Success);
        }

        break;
    }

    return Success;
}
/*===========================================================================*/
static Bool PenmountConvert(InputInfoPtr pInfo, int first, int num,
            int v0, int v1, int v2, int v3, int v4, int v5, int *x, int *y)
{
    PenmountPtr pPenmount = pInfo->private;
    int screenWidth = screenInfo.screens[pPenmount->screen]->width;
    int screenHeight = screenInfo.screens[pPenmount->screen]->height;

    /*xf86Msg(X_INFO, "penmount: Convert\n");*/

    if (first != 0 || num != 2)
        return FALSE;

    *x = (v0 - pPenmount->min_x) * screenWidth / (pPenmount->max_x - pPenmount->min_x);
    *y = (v1 - pPenmount->min_y) * screenHeight / (pPenmount->max_y - pPenmount->min_y);

    return TRUE;
}
/*===========================================================================*/
static int PenmountProbe(InputInfoPtr pInfo)
{
    PenmountPtr pPenmount = pInfo->private;

    /*xf86Msg(X_INFO, "penmount: Probe\n");*/

    if (pPenmount->chip == PM6000USB ||
        pPenmount->chip == PM5000    ||
        pPenmount->chip == PMLPC) {

        if (ioctl(pInfo->fd, EVIOCGRAB, (void *)1) && errno == EINVAL) {
            /* keyboards are unsafe in 2.4 */
            pPenmount->kernel24 = 1;
        } else {
            ioctl(pInfo->fd, EVIOCGRAB, (void *)0);
            pPenmount->kernel24 = 0;
        }
    }
    
    return 0;
}

/*===========================================================================*/
static InputInfoPtr PenmountPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
{
    InputInfoPtr    pInfo;
    const char      *device;
    PenmountPtr     pPenmount;
    int             fd;
    int             i;
    char            devnode[32];
    struct input_id id;
    char            *s;

    xf86Msg(X_INFO, "penmount: PreInit\n");

    if (!(pInfo = xf86AllocateInput(drv, 0)))
        return NULL;

    /* Initialise the InputInfoRec. */
    pInfo->name = dev->identifier;
    pInfo->flags = 0;
    pInfo->type_name = XI_TOUCHSCREEN;
    pInfo->device_control = PenmountProc;
    pInfo->read_input = PenmountReadInput;
    pInfo->history_size = 0;
    pInfo->control_proc = NULL;
    pInfo->close_proc = NULL;
    pInfo->switch_mode = NULL;
    pInfo->conversion_proc = PenmountConvert;
    pInfo->reverse_conversion_proc = NULL;
    pInfo->dev = NULL;
    pInfo->private_flags = 0;
    pInfo->always_core_feedback = 0;
    pInfo->conf_idev = dev;
    pInfo->flags = XI86_POINTER_CAPABLE | XI86_SEND_DRAG_EVENTS;

/*
    if (!(pPenmount = xcalloc(sizeof(*pPenmount), 1)))
        return pInfo;
    pInfo->private = pPenmount;
*/
    if (!(pPenmount = xcalloc(1, sizeof(PenmountRec))))
        return pInfo;

    pInfo->private = pPenmount;

    xf86CollectInputOptions(pInfo, penmountDefaults, NULL);
    xf86ProcessCommonOptions(pInfo, pInfo->options);
/*    xf86OptionListReport(pInfo->options);*/

    pPenmount->screen       = xf86SetIntOption(pInfo->options, "ScreenNumber", 0);

    pPenmount->xmin_offset  = xf86SetIntOption(pInfo->options, "XMinOffset" , 10);
    pPenmount->xmax_offset  = xf86SetIntOption(pInfo->options, "XMaxOffset" , 10);
    pPenmount->ymin_offset  = xf86SetIntOption(pInfo->options, "YMinOffset" , 10);
    pPenmount->ymax_offset  = xf86SetIntOption(pInfo->options, "YMaxOffset" , 10);

    pPenmount->std_xmin     = xf86SetIntOption(pInfo->options, "StdXMin" , 20);
    pPenmount->std_xmax     = xf86SetIntOption(pInfo->options, "StdXMax" , 2000);
    pPenmount->std_ymin     = xf86SetIntOption(pInfo->options, "StdYMin" , 20);
    pPenmount->std_ymax     = xf86SetIntOption(pInfo->options, "StdYMax" , 2000);

    pPenmount->screen_scale = xf86SetIntOption(pInfo->options, "ScreenScale" , 0);
    pPenmount->rotation     = xf86SetIntOption(pInfo->options, "Rotation" , 1);

    pPenmount->average_point = xf86SetIntOption(pInfo->options, "AveragePoint" , 0);

    pPenmount->beep         = xf86SetIntOption(pInfo->options, "Beep", 1);
    pPenmount->press_vol    = xf86SetIntOption(pInfo->options, "PressVol", 100);
    pPenmount->press_pitch  = xf86SetIntOption(pInfo->options, "PressPitch", 880);
    pPenmount->press_dur    = xf86SetIntOption(pInfo->options, "PressDur", 15) & 0xff;
    pPenmount->rel_vol      = xf86SetIntOption(pInfo->options, "ReleaseVol", 0);
    pPenmount->rel_pitch    = xf86SetIntOption(pInfo->options, "ReleasePitch", 1200);
    pPenmount->rel_dur      = xf86SetIntOption(pInfo->options, "ReleaseDur", 10) & 0xff;
    
    pPenmount->hold_time    = xf86SetIntOption(pInfo->options, "HoldTime", 500000);
    pPenmount->calib_hold_time = xf86SetIntOption(pInfo->options, "CalibHoldTime", 1500000);
    pPenmount->lock_window_range = xf86SetIntOption(pInfo->options, "LockWindowRange" , 64);

    pPenmount->button2      = xf86SetIntOption(pInfo->options, "Button2" , 3);/* 1=Left, 2=Middle, 3=Right*/
    pPenmount->pendown_mode = xf86SetIntOption( pInfo->options, "PenDownMode", 0);/* 0:stream mode, 1:point mode */
    pPenmount->ad_bit       = xf86SetIntOption( pInfo->options, "ADBit", 10 );
    
    pPenmount->buffer       = NULL;
    pPenmount->lex_mode     = PENMOUNT_BYTE0;
    pPenmount->old_pendown  = 0;
    pPenmount->pendown      = 0;

	s = xf86FindOptionValue (pInfo->options, "ReportingMode");
	if ((s) && (xf86NameCmp (s, "raw") == 0))
		pPenmount->reporting_mode = TS_Raw;
	else
		pPenmount->reporting_mode = TS_Scaled;

    s = xf86SetStrOption(pInfo->options, "ConfigFile", NULL);
    if (s)
        sprintf(pPenmount->configfile, "%s", s);
    else
        sprintf(pPenmount->configfile, "/etc/penmount.dat");

    device = xf86CheckStrOption(dev->commonOptions, "Path", NULL);
    if (!device)
        device = xf86CheckStrOption(dev->commonOptions, "Device", NULL);
    if (!device) {
        xf86Msg(X_ERROR, "%s: No device specified.\n", pInfo->name);
        xf86DeleteInput(pInfo, 0);
        return NULL;
    }

    /* Penmount serial ? */
    pPenmount->chip = UNKNOWN_CHIP;
    s = xf86FindOptionValue (pInfo->options, "Protocol");
    if (s) {
        if (xf86NameCmp (s, "PM9000") == 0) {
            pPenmount->chip = PM9000;
            pInfo->read_input = Penmount232ReadInput;
            pInfo->control_proc = PenmountControlProc;
            pInfo->close_proc = PenmountCloseProc;
            pInfo->switch_mode = PenmountSwitchMode;
            pInfo->conversion_proc = PenmountConvertProc;
            xf86Msg(X_INFO, "penmount chip : 9000\n");
        } else if (xf86NameCmp (s, "PM6000COM") == 0) {
            pPenmount->chip = PM6000COM;
            pInfo->read_input = Penmount232ReadInput;
            pInfo->control_proc = PenmountControlProc;
            pInfo->close_proc = PenmountCloseProc;
            pInfo->switch_mode = PenmountSwitchMode;
            pInfo->conversion_proc = PenmountConvertProc;
            xf86Msg(X_INFO, "penmount chip : 6000\n");
        }
    }

    /* detect Penmount USB series */
    if (pPenmount->chip == UNKNOWN_CHIP) {
        /* look for penmount device node */
        for (i=0; i<32; i++) {
            sprintf(devnode, "/dev/input/event%d", i);
            do {
                fd = open(devnode, O_RDWR, 0);
            } while (fd < 0 && errno == EINTR);
        
            if (fd < 0) {
                xf86Msg(X_ERROR, "open %s error!\n", devnode);
                continue;
            }
            ioctl(fd, EVIOCGID, &id);
            if ((id.vendor == 0x14e1 && id.product == 0x6000) ||
                (id.vendor == 0x14e1 && id.product == 0x5000) ||
                (id.vendor == 0x1204 && id.product == 0x9998) ||
                (id.vendor == 0x14e1 && id.product == 0xfbfb)) {
                device = (const char*)devnode;
                xf86Msg(X_INFO, "PenMount found! (%04x:%04x)\n", id.vendor, id.product);

                if (id.vendor == 0x14e1 && id.product == 0x6000)
                    pPenmount->chip = PM6000USB;
                else if (id.vendor == 0x14e1 && id.product == 0x5000)
                    pPenmount->chip = PM5000;
                else if (id.vendor == 0x1204 && id.product == 0x9998)
                    pPenmount->chip = PM5000;
                else if (id.vendor == 0x14e1 && id.product == 0xfbfb)
                    pPenmount->chip = PMLPC;

                close(fd);
                break;
            }
            close(fd);
        }

        do {
            pInfo->fd = open(device, O_RDWR, 0);
        } while (pInfo->fd < 0 && errno == EINTR);

        if (pInfo->fd < 0) {
            xf86Msg(X_ERROR, "Unable to open penmount device \"%s\".\n", device);
            xf86DeleteInput(pInfo, 0);
            xfree(pPenmount);
            return NULL;
        }
    } else if (pPenmount->chip == PM9000 || pPenmount->chip == PM6000COM) {
        pInfo->fd = xf86OpenSerial (pInfo->options);
        if (pInfo->fd == -1) {
            xf86Msg(X_ERROR, "penmount: Unable to open penmount device (com port).\n");
            goto SetupProc_fail;
        }
        xf86CloseSerial(pInfo->fd);

    } else { /* can not find any Penmount */
SetupProc_fail:
        xf86DeleteInput(pInfo, 0);
        xfree(pPenmount);
        return NULL;
    }


    pPenmount->calib_mode = SetCalibration(pPenmount->configfile, 1);
    if (pPenmount->calib_mode == 0)
        LoadStdCalib(pPenmount);
    
    pInfo->flags |= XI86_CONFIGURED;

    if (PenmountProbe(pInfo)) {
        xf86DeleteInput(pInfo, 0);
        return NULL;
    }

    return pInfo;
}
/*===========================================================================*/
_X_EXPORT InputDriverRec PENMOUNT = {
    1,
    "penmount",
    NULL,
    PenmountPreInit,
    NULL,
    NULL,
    0
};
/*===========================================================================*/
static void PenmountUnplug(pointer p)
{
}
/*===========================================================================*/
static pointer PenmountPlug(pointer module, pointer options, int *errmaj, int *errmin)
{
    xf86AddInputDriver(&PENMOUNT, module, 0);
    return module;
}
/*===========================================================================*/
static XF86ModuleVersionInfo PenmountVersionRec =
{
    "penmount",
    MODULEVENDORSTRING,
    MODINFOSTRING1,
    MODINFOSTRING2,
    0, /* Missing from SDK: XORG_VERSION_CURRENT, */
    1, 0, 0,
    ABI_CLASS_XINPUT,
    ABI_XINPUT_VERSION,
    MOD_CLASS_XINPUT,
    {0, 0, 0, 0}
};
/*===========================================================================*/
_X_EXPORT XF86ModuleData penmountModuleData =
{
    &PenmountVersionRec,
    PenmountPlug,
    PenmountUnplug
};
